#include <conio.h>
#include "sin.h"
#include "cos.h"

#define CORRECT       54613L

extern "C" TRIANGLE(unsigned x1, unsigned y1,
                    unsigned x2, unsigned y2,
                    unsigned x3, unsigned y3,
                    unsigned color, unsigned page);

extern "C" long SQRT(long arg);

unsigned page = 0;

class TPoint
{
  public:
    int x, y, z;
    TPoint() {}
    TPoint(long x, long y, long z)
    {
      TPoint::x = x;
      TPoint::y = y;
      TPoint::z = z;
    }
    void CorrectAspect()
    {
      y = int((y * CORRECT) >> 16);
    }
};

TPoint RotateX(TPoint p, unsigned angle)
{
  return TPoint(p.x,
                (p.y * cos_table[angle] - p.z * sin_table[angle]) >> 16,
                (p.y * sin_table[angle] + p.z * cos_table[angle]) >> 16);
}

TPoint RotateY(TPoint p, unsigned angle)
{
  return TPoint((p.x * cos_table[angle] - p.z * sin_table[angle]) >> 16,
                p.y,
                (p.x * sin_table[angle] + p.z * cos_table[angle]) >> 16);
}

TPoint RotateZ(TPoint p, unsigned angle)
{
  return TPoint((p.x * cos_table[angle] - p.y * sin_table[angle]) >> 16,
                (p.x * sin_table[angle] + p.y * cos_table[angle]) >> 16,
                p.z);
}

class TNormal
{
    long x, y, z;
  public:
    TNormal() {}
    TNormal(long x, long y, long z)
    {
      TNormal::x = x;
      TNormal::y = y;
      TNormal::z = z;
    }
    TNormal(TPoint p1, TPoint p2, TPoint p3)
    {
      long register t;
      TNormal n1(p2.x - p1.x, p2.y - p1.y, p2.z - p1.z),
              n2(p3.x - p1.x, p3.y - p1.y, p3.z - p1.z);

      x = n1.y * n2.z - n1.z * n2.y;
      y = n1.z * n2.x - n1.x * n2.z;
      z = n1.x * n2.y - n1.y * n2.x;

      t = (long)SQRT(x * x + y * y + z * z);

      if (t)
      {
        x = (x << 16) / t;
        y = (y << 16) / t;
        z = (z << 16) / t;
      }
      else
      {
        x <<= 16;
        y <<= 16;
        z <<= 16;
      }
    }
    long Z()
    {
      return z;
    }
};

void Triangle2D(TPoint p1, TPoint p2, TPoint p3,
                unsigned x, unsigned y, unsigned color)
{
  TRIANGLE(p1.x + x, p1.y + y,
           p2.x + x, p2.y + y,
           p3.x + x, p3.y + y,
           color, page);
}

class TTriangle {
    TPoint m1, m2, m3;
    int light;
  public:
    TNormal n;
    TTriangle(): n(), m1(), m2(), m3() {}
    TTriangle(TPoint p1, TPoint p2, TPoint p3): n(p1, p2, p3), m1(p1), m2(p2), m3(p3)
    {
      light = int(((n.Z() << 6) - 1) >> 16);
    }
    TTriangle(TPoint p1, TPoint p2, TPoint p3, TNormal n_): n(n_), m1(p1), m2(p2), m3(p3)
    {
      light = int(((n.Z() << 6) - 1) >> 16);
    }
    int Visible()
    {
      return (n.Z() > 0);
    }
    void Draw(unsigned color, unsigned x, unsigned y)
    {
      Triangle2D(m1, m2, m3, x, y, color);
    }
    int Light()
    {
      return light;
    }
    void SetLight(int light)
    {
      if (light > 0)
        TTriangle::light = light;
    }
};

class TDodecaedr {
    TPoint p[20], p2[20];
    TTriangle t[36];
    int s;
  public:
    TDodecaedr(int size);
    void Draw(unsigned, unsigned, unsigned, unsigned, unsigned, unsigned);
};

TDodecaedr::TDodecaedr(int size)
{
  for (int i = 0; i < 10; i++)
  {
    float k = (1 - (i % 2) * 2) * 0.2 * (float)size;
    p[i] = TPoint((-cos_table[((i + 1) * 36) % 360] * size) >> 16, (-sin_table[((i + 1) * 36) % 360] * size) >> 16, k);
  }

  for (int i = 0; i < 5; i++)
  {
    p[i + 10] = TPoint(long(-(float)cos_table[i * 72 + 36] * 0.60 * (float)size) >> 16, long(-(float)sin_table[i * 72 + 36] * 0.6 * (float)size) >> 16, 0.8 * (float)size);
    p[i + 15] = TPoint(long(-(float)cos_table[i * 72] * 0.60 * (float)size) >> 16, long(-(float)sin_table[i * 72] * 0.6 * (float)size) >> 16, -0.8 * (float)size);
  }
}

void TDodecaedr::Draw(unsigned angle_x, unsigned angle_y,
                  unsigned angle_z, unsigned x,
                  unsigned y, unsigned color)
{
  int i = 0, offs;

  for (; i < 20; i++)
  {
    p2[i] = TPoint(RotateX(RotateY(RotateZ(p[i], angle_z), angle_y), angle_x));
    p2[i].CorrectAspect();
  }

  for (i = 0; i < 5; i++)
  {
    t[i] = TTriangle(p2[10 + i], p2[i << 1], p2[(i << 1) + 1]);
    t[i + 5] = TTriangle(p2[10 + i], p2[(i << 1) + 1], p2[((i + 1) % 5) + 10], t[i].n);
    t[i + 10] = TTriangle(p2[((i + 1) % 5) + 10], p2[(i << 1) + 1], p2[((i << 1) + 2) % 10], t[i].n);
    t[i + 15] = TTriangle(p2[((i << 1) + 2) % 10], p2[(i << 1) + 1], p2[((i + 1) % 5) + 15]);
    t[i + 20] = TTriangle(p2[((i + 2) % 5) + 15], p2[((i << 1) + 2) % 10], p2[((i + 1) % 5) + 15], t[i + 15].n);
    t[i + 25] = TTriangle(p2[((i << 1) + 3) % 10], p2[((i << 1) + 2) % 10], p2[((i + 2) % 5) + 15], t[i + 15].n);
  }
  t[30] = TTriangle(p2[10], p2[11], p2[12]);
  t[31] = TTriangle(p2[10], p2[12], p2[14], t[30].n);
  t[32] = TTriangle(p2[14], p2[12], p2[13], t[30].n);
  t[33] = TTriangle(p2[17], p2[16], p2[15]);
  t[34] = TTriangle(p2[19], p2[17], p2[15], t[33].n);
  t[35] = TTriangle(p2[18], p2[17], p2[19], t[33].n);

  for (i = 0; i < 36; i++)
  {
    if (t[i].Visible())
    {
      offs = t[i].Light();
      t[i].Draw((color << 6) + offs, x, y);
    }
  }
}

class TFlyer {
    int x, y;
    int a_x, a_y, a_z;
    unsigned size, color;
    int dx, dy, d_ax, d_ay, d_az;
    TDodecaedr t;
  public:
    TFlyer(unsigned size, int dx, int dy, int d_ax, int d_ay, int d_az,
           unsigned color):
          t(size)
    {
      TFlyer::size = size;
      TFlyer::color = color;
      TFlyer::dx = dx;
      TFlyer::dy = dy;
      TFlyer::d_ax = d_ax;
      TFlyer::d_ay = d_ay;
      TFlyer::d_az = d_az;
      a_x = a_y = a_z = 0;
      x = 160;
      y = 100;
    }
    void Move()
    {
      t.Draw(a_x, a_y, a_z, x, y, color);

      x += dx;
      y += dy;

      a_x += d_ax;
      a_y += d_ay;
      a_z += d_az;

      if ((x > 320 - size - size / 2) || (x < size + size / 2)) dx = -dx;
      if ((y > 200 - size - size / 2) || (y < size + size / 2)) dy = -dy;

      if (a_x < 0) a_x = 360 + a_x;
      if (a_x > 359) a_x = 0;
      if (a_y < 0) a_y = 360 + a_y;
      if (a_y > 359) a_y = 0;
      if (a_z < 0) a_z = 360 + a_z;
      if (a_z > 359) a_z = 0;
    }
};

void SetGraphMode()
{
  asm  mov    ax,  13h
  asm  int    10h
  asm  mov    dx,  3C4h
  asm  mov    al,  4
  asm  out    dx,  al
  asm  inc    dx
  asm  in     al,  dx
  asm  or     al,  4
  asm  and    al,  0F7h
  asm  out    dx,  al
  asm  mov    dx,  3D4h
  asm  mov    al,  14h
  asm  out    dx,  al
  asm  inc    dx
  asm  in     al,  dx
  asm  and    al,  0BFh
  asm  out    dx,  al
  asm  dec    dx
  asm  mov    al,  17h
  asm  out    dx,  al
  asm  inc    dx
  asm  in     al,  dx
  asm  or     al,  40h
  asm  out    dx,  al
  asm  push   0A000h
  asm  pop    es
  asm  xor    di,  di
  asm  xor    ax,  ax
  asm  mov    cx,  8000h
  asm  cld
  asm  rep    stosw
}

void SetTextMode()
{
  asm  mov    ax,  03h
  asm  int    10h
}

void ClearPage(unsigned page)
{
  asm  push   0A000h
  asm  pop    es
  asm  mov    di,  page
  asm  xor    ax,  ax
  asm  mov    cx,  8192
  asm  mov    dx,  3C4h
  asm  mov    ax,  0F02h
  asm  out    dx,  ax
  asm  cld
  asm  mov    cx,  8192
  asm  xor    ax,  ax
  asm  rep    stosw
}

void TilePage(unsigned page, char color)
{
  asm  push   0A000h
  asm  pop    es
  asm  mov    di,  page
  asm  mov    dx,  3C4h
  asm  mov    ax,  0F02h
  asm  out    dx,  ax
  asm  mov    cx,  50
  asm  mov    bl,  color
  asm  cld
l:asm  push   cx
  asm  mov    al,  bl
  asm  add    al,  cl
  asm  mov    ah,  al
  asm  mov    cx,  160
  asm  rep    stosw
  asm  pop    cx
  asm  loop   l
}

void SetPage(unsigned page)
{
  asm  mov    bx,  page
  asm  mov    dx,  3D4h
  asm  mov    al,  0Ch
  asm  mov    ah,  bh
  asm  out    dx,  ax
  asm  mov    al,  0Dh
  asm  xor    ah,  ah
  asm  out    dx,  ax
}

void Vertret()
{
    asm  mov    dx,  03DAh
l1: asm  in     al,  dx
    asm  test   al,  8
    asm  jz     l1
l2: asm  in     al,  dx
    asm  test   al,  8
    asm  jnz    l2
}

void SetColor(char color, char r, char g, char b)
{
  asm  mov    dx,  3C8h
  asm  mov    al,  color
  asm  out    dx,  al
  asm  inc    dx
  asm  mov    al,  r
  asm  out    dx,  al
  asm  mov    al,  g
  asm  out    dx,  al
  asm  mov    al,  b
  asm  out    dx,  al
}

void SetPalette()
{
  for (int i = 0; i < 64; i++)
  {
    SetColor(i, 0, 0, i);
    SetColor(i + 64, 0, i, 0);
    SetColor(i + 128, i, 0, 0);
    SetColor(i + 192, i, i, 0);
  }
}

void Message()
{
  cprintf("3D transformations - by HackMaster\n");
}

void main()
{
  SetGraphMode();
  SetPalette();
  TFlyer f1(80, 0, 0, 2, 1, 2, 0);

  do
  {
    SetPage(page ^ 0x4000);
    Vertret();
    TilePage(page, 0x80);
    f1.Move();
    page ^= 0x4000;
  } while (!kbhit());

  SetTextMode();
  Message();
}
